
from sklearn.linear_model import LinearRegression
from sklearn.linear_model import LogisticRegression # for classModel


from sklearn.tree import DecisionTreeRegressor
from sklearn.tree import DecisionTreeClassifier

from sklearn.neighbors import KNeighborsClassifier

from sklearn.ensemble import BaggingRegressor
from sklearn.ensemble import RandomForestRegressor
from sklearn.ensemble import BaggingClassifier
from sklearn.ensemble import RandomForestClassifier

from sklearn.pipeline import make_pipeline

from .evaluate_regression_file import evaluate_regression
from .evaluate_classification_file import evaluate_classification



def modelSearch(preprocessor, X_train, y_train, X_test, y_test, model_type="num", random_state=42):
    if model_type == 'num':
        lin_reg = LinearRegression()
        dec_tree = DecisionTreeRegressor(random_state=random_state)
        bagreg = BaggingRegressor(random_state=random_state)
        rf = RandomForestRegressor(random_state=random_state)

        arrNum = [lin_reg, dec_tree, bagreg, rf]

        for model in arrNum:
            model_name = model.__class__.__name__
            model_selected_pip = make_pipeline(preprocessor, model, memory=model_name)
            model_selected_pip.fit(X_train, y_train)
            print("\n")
            print("\n")
            print(":" * 50)
            print("Model Name:", model_name)
            evaluate_regression(model_selected_pip, X_train, y_train, X_test, y_test)
            print(":" * 50)
            print("\n")
            print("\n")

    elif model_type == 'class':
        dec_tree = DecisionTreeClassifier(random_state=random_state)
        knn = KNeighborsClassifier()
        logreg = LogisticRegression(random_state=random_state, C=1000)
        bagregClass = BaggingClassifier(random_state=random_state)
        rfc = RandomForestClassifier(random_state=random_state)

        arrClass = [dec_tree, knn, logreg, bagregClass, rfc]

        for model in arrClass:
            model_name = model.__class__.__name__
            model_selected_pip = make_pipeline(preprocessor, model, memory=model_name)
            model_selected_pip.fit(X_train, y_train)
            print("\n")
            print("\n")
            print(":" * 50)
            print("Model Name:", model_name)
            evaluate_classification(model_selected_pip, X_train, y_train, X_test, y_test)
            print(":" * 50)
            print("\n")
            print("\n")




import plotly.express as px
import plotly.io as pio


import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from sklearn.decomposition import PCA
from sklearn import set_config
set_config(transform_output='pandas')
pd.set_option('display.max_columns',100)


def update_scatter3d(fig):
    fig.update_traces({'marker':{'size':3}})
    fig.show(config={'scrollZoom':False})



def plot_PCA(preprocessor, X_train, y_train, X_test, y_test, color_target='target', pairplot = False):


  X = pd.concat([X_train,X_test], axis=0)

  y = pd.concat([y_train,y_test], axis=0)


  X_tf = preprocessor.fit_transform(X)



  # Instantiate PCA to make 3 principal components
  pca = PCA(n_components=3)
  # Create and define the principal components
  principal_components = pca.fit_transform(X_tf)


  # Concatenate principal components with target
  plot_df_pca = pd.concat([principal_components, y], axis=1)

  if pairplot == True:
    # Concatenate principal components with target
    plot_df_pca = pd.concat([principal_components, y], axis=1)
    # Plot with color coding based on target
    g_pca = sns.pairplot(data=plot_df_pca,  vars=principal_components.columns, hue='Training')
    g_pca.fig.suptitle('Visualizing First 3 PCs - Colored by Training', y=1.01);

  # Make a 3d scatter plot with a PC on each axis and color by the target
  # Change template style to plotly_dark
  fig = px.scatter_3d(plot_df_pca, x='pca0',y='pca1',z='pca2', width=800, height=600, color = color_target, template = 'plotly_dark')
  update_scatter3d(fig)

  return fig, pca



from sklearn.metrics import classification_report, ConfusionMatrixDisplay
import matplotlib.pyplot as plt

def classification_metrics(y_true, y_pred, label='',
                           output_dict=False, figsize=(8,4),
                           normalize='true', cmap='Blues',
                           colorbar=False):
  # Get the classification report
  report = classification_report(y_true, y_pred)
  ## Print header and report
  header = "-"*70
  print(header, f" Classification Metrics: {label}", header, sep='\n')
  print(report)
  ## CONFUSION MATRICES SUBPLOTS
  fig, axes = plt.subplots(ncols=2, figsize=figsize)
  # create a confusion matrix  of raw counts
  ConfusionMatrixDisplay.from_predictions(y_true, y_pred,
                normalize=None, cmap='gist_gray', colorbar=colorbar,
                ax = axes[0],);
  axes[0].set_title("Raw Counts")
  # create a confusion matrix with the test data
  ConfusionMatrixDisplay.from_predictions(y_true, y_pred,
                normalize=normalize, cmap=cmap, colorbar=colorbar,
                ax = axes[1]);
  axes[1].set_title("Normalized Confusion Matrix")
  # Adjust layout and show figure
  fig.tight_layout()
  plt.show()
  # Return dictionary of classification_report
  if output_dict==True:
    report_dict = classification_report(y_true, y_pred, output_dict=True)
    return report_dict



def evaluate_classification(model, X_train, y_train, X_test, y_test,
                         figsize=(6,4), normalize='true', output_dict = False,
                            cmap_train='Blues', cmap_test="Reds",colorbar=False):
  # Get predictions for training data
  y_train_pred = model.predict(X_train)
  # Call the helper function to obtain regression metrics for training data
  results_train = classification_metrics(y_train, y_train_pred, #verbose = verbose,
                                     output_dict=True, figsize=figsize,
                                         colorbar=colorbar, cmap=cmap_train,
                                     label='Training Data')
  print()
  # Get predictions for test data
  y_test_pred = model.predict(X_test)
  # Call the helper function to obtain regression metrics for test data
  results_test = classification_metrics(y_test, y_test_pred, #verbose = verbose,
                                  output_dict=True,figsize=figsize,
                                         colorbar=colorbar, cmap=cmap_test,
                                    label='Test Data' )
  if output_dict == True:
    # Store results in a dataframe if ouput_frame is True
    results_dict = {'train':results_train,
                    'test': results_test}
    return results_dict



from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score
import matplotlib.pyplot as plt

def KMeansCoff(scaled_df):
  # Iterate over a range of k clusters to find the optimal number of clusters
  ks = range(2,11)
  sils = []
  inertias = []
  for k in ks:
    kmeans = KMeans(n_clusters=k, n_init = 'auto', random_state=42)
    kmeans.fit(scaled_df)
    sils.append(silhouette_score(scaled_df, kmeans.labels_))
    inertias.append(kmeans.inertia_)
  #plot inertias and silhouette scores for each number of clusters.
  fig, axes = plt.subplots(1,2, figsize=(15,5))
  axes[0].plot(ks, sils)
  axes[0].set_title('Silhouette Scores')
  axes[0].set_xticks(ks)
  axes[1].plot(ks, inertias)
  axes[1].set_title('Inertia')
  axes[1].set_xticks(ks);


import matplotlib.pyplot as plt
import seaborn as sns

# Updating function
def plot_categorical_vs_target(df, x, y, figsize=(6,4),
                            fillna = True, placeholder = 'MISSING',
                            order = None, target_type='reg'):
  # Make a copy of the dataframe and fillna
  temp_df = df.copy()
  # fillna with placeholder
  if fillna == True:
    temp_df[x] = temp_df[x].fillna(placeholder)
  # or drop nulls prevent unwanted 'nan' group in stripplot
  else:
    temp_df = temp_df.dropna(subset=[x])
  # Create the figure and subplots
  fig, ax = plt.subplots(figsize=figsize)
  # REGRESSION-TARGET PLOT
  if target_type=='reg': #Added if statement here
    # Barplot
    sns.barplot(data=temp_df, x=x, y=y, ax=ax, order=order, alpha=0.6,
                linewidth=1, edgecolor='black', errorbar=None)
    # Boxplot
    sns.stripplot(data=temp_df, x=x, y=y, hue=x, ax=ax,
                  order=order, hue_order=order, legend=False,
                  edgecolor='white', linewidth=0.5,
                  size=3,zorder=0)
  # CLASSIFICATION-TARGET PLOT # This is the new code for the classification task
  elif target_type=='class':
    ax = sns.histplot(data=temp_df, hue=y, x=x, stat='percent',  multiple='fill')
  # Rotate xlabels
  ax.set_xticks(ax.get_xticks()) # Added this to prevent a bug
  ax.set_xticklabels(ax.get_xticklabels(), rotation=45, ha='right')
  # Add a title
  ax.set_title(f"{x} vs. {y}")
  fig.tight_layout()
  return fig, ax






from .preprocessor_maker_file import preprocessor_maker
from .neural_networks_builder_file import neural_networks_builder
from .KMeansCoff_file import KMeansCoff
from .evaluate_classification_file import evaluate_classification
from .evaluate_regression_file import evaluate_regression
from .explore_categorical_file import explore_categorical
from .plot_categorical_vs_target_file import plot_categorical_vs_target
from .explore_numeric_file import explore_numeric
from .modelSearch_file import modelSearch
from .plot_numeric_vs_target_file import plot_numeric_vs_target
from .plot_PCA_file import plot_PCA
# from .all_imports_file import all_imports


from sklearn.model_selection import train_test_split
from sklearn.impute import SimpleImputer
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler, OneHotEncoder, OrdinalEncoder
from sklearn.compose import ColumnTransformer
from sklearn.compose import make_column_selector
# set the default output to pandas
from sklearn import set_config
set_config(transform_output='pandas')


def preprocessor_maker(df, target, num_imputer='mean', fill_num = 0, ord_features=[], ord_orders=[]
          , ord_imputer='constant', fill_ord='NA', cat_imputer='constant'
          ,fill_cat='NA',  random_state=42, num_scale = True, ord_scale = True, cat_scale = True ):


  y = df[target]
  X = df.drop([target] , axis=1)
  X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=random_state)

  indicator_arr = [0,0,0]

  if (len(X_train.select_dtypes('number').columns)):
    num_selector = make_column_selector(dtype_include='number')

    if num_imputer == 'constant':
      num_imputer1 = SimpleImputer(strategy='constant', fill_value = fill_num)
    else:
      num_imputer1 = SimpleImputer(strategy=num_imputer)

    if num_scale == True:
      scaler = StandardScaler()
      num_pipe = make_pipeline(num_imputer1, scaler)
    else:
      num_pipe = make_pipeline(num_imputer1)


    num_tuple = ('numeric',num_pipe, num_selector)

    indicator_arr[0] = 1



  # ================================================
  if len(ord_features):

    ord_cols = ord_features


    if ord_imputer == 'constant':
      impute_na_ord = SimpleImputer(strategy='constant', fill_value=fill_ord)
    else:
      impute_na_ord = SimpleImputer(strategy=ord_imputer)


    ord_encoder = OrdinalEncoder(categories=ord_orders)

    # # Making a final scaler to scale category #'s
    # scaler_ord = StandardScaler()

    if ord_scale == True:
      scaler_ord = StandardScaler()
      ord_pipe = make_pipeline(impute_na_ord, ord_encoder, scaler_ord)
    else:
      ord_pipe = make_pipeline(impute_na_ord, ord_encoder)

    # ord_pipe = make_pipeline(impute_na_ord, ord_encoder, scaler_ord)



    ord_tuple = ('ordinal', ord_pipe, ord_cols)
    # indicator_arr.append(ord_tuple)
    indicator_arr[1] = 1
  # ==================================================================

  # (New) Select columns with make_column_selector----------------------------------------------
  if (len(X_train.select_dtypes('object').columns)):
    # cat_selector = make_column_selector(dtype_include='object')

    if len(ord_features):
      ohe_cols = X_train.select_dtypes('object').drop(columns=ord_cols).columns


    else:
      ohe_cols = X_train.select_dtypes('object').columns


    if cat_imputer == 'constant':
      freq_imputer = SimpleImputer(strategy='constant', fill_value=fill_cat)
    else:
      freq_imputer = SimpleImputer(strategy=cat_imputer)
      
    ohe = OneHotEncoder(sparse_output=False, handle_unknown='ignore')


    if cat_scale == True:
      scaler_cat = StandardScaler()
      cat_pipe = make_pipeline(freq_imputer, ohe, scaler_cat)
    else:
      cat_pipe = make_pipeline(freq_imputer, ohe)


    cat_tuple = ('categorical',cat_pipe, ohe_cols)
    indicator_arr[2] = 1


  # =======================================================================


  if indicator_arr[0] == 1 and indicator_arr[1] == 1 and indicator_arr[2] == 1:
    all_tuple = [num_tuple, ord_tuple, cat_tuple]

  elif indicator_arr[0] == 1 and indicator_arr[1] == 1 and indicator_arr[2] == 0:
    all_tuple = [num_tuple, ord_tuple]

  elif indicator_arr[0] == 1 and indicator_arr[1] == 0 and indicator_arr[2] == 1:
    all_tuple = [num_tuple, cat_tuple]

  elif indicator_arr[0] == 0 and indicator_arr[1] == 1 and indicator_arr[2] == 1:
    all_tuple = [ord_tuple, cat_tuple]

  elif indicator_arr[0] == 1 and indicator_arr[1] == 0 and indicator_arr[2] == 0:
    all_tuple = [num_tuple]

  elif indicator_arr[0] == 0 and indicator_arr[1] == 0 and indicator_arr[2] == 1:
    all_tuple = [cat_tuple]

  elif indicator_arr[0] == 0 and indicator_arr[1] == 1 and indicator_arr[2] == 0:
    all_tuple = [ord_tuple]

  else:
    print('Error')



  preprocessor = ColumnTransformer(all_tuple, verbose_feature_names_out=False)

  return preprocessor, X_train, y_train, X_test, y_test


